// University of Washington Bothell Graphics Library
// Authors: Kelvin Sung, Steve Baer
// The accompanying library supports CSS Graphics courses taught at UW-Bothell
// See: http://courses.washington.edu/css450/
//      http://courses.washington.edu/css451/
///////////////////////////////////////////////////////////////////////////////////
#pragma once
#include "uwbgl_Common.h"
#include "uwbgl_DrawHelper5.h"
#include "uwbgl_Intersect5.h"
#include "uwbgl_Material1.h"

/// Base class for all graphics objects
/// A generic primitive represents something that can be drawn with given attributes
/// in a graphics window. It can aslo pay attention and respond to mouse input events.
/// For the most part, primitives are typically 2D or 3D geometry that represents
/// different shapes. This is not a necessary requirement, but is the common use. Primitives
/// could also be used to display non-interactive items like heads up display and text
class UWB_Primitive
{
public:
  UWB_Primitive();
  virtual ~UWB_Primitive();

  /// Draw this primitive to a graphics window. This function just makes sure that the
  /// primitive is visible and then calls the virtual SetupDrawAttributes and DrawPrimitive
  /// functions
  /// \param lod requested level of detail for drawing this privitive. Many primitive classed
  ///        may ignore this parameter, but some primitives like curves and circles would use
  ///        this to determine how many straight segments would be used to represent a curved primitive
  /// \param draw_helper this object contains generic drawing functions that may be called by
  ///        a primitive to help abstract API specific drawing calls from this primitive object
  void Draw( eLevelOfDetail lod, UWB_DrawHelper& draw_helper );

  /// Update the state of a primitive that may have changed over a period of time. This may be used
  /// to adjust things like position or an attribute that may change with time
  /// \param elapsed_seconds amount of time that has passed since the last call to Update
  virtual void Update(float elapsed_seconds);

  /// Event handler for mouse input events
  /// \param vertexID used by applications to define which mouse event this relates to.
  ///        Typically the first mouse down event would correspond to a vertexID of 0 and
  ///        subsequent calls would pass a vertxID of 1
  /// \param x,y coordinates of mouse position
  virtual void MouseDownVertex(int vertexID, float x, float y);

  /// Virtual function for moving the position of a primitive object
  /// \param x, y position to mose primitive to
  virtual void MoveTo(float x, float y);

  /// Virtual function to get position of primitive in 3D coordinates
  /// \return the object's position
  virtual vec3 GetLocation() const;

  /// Get the bounding volume of this object
  /// This is used to determine if this primitive's bounds intersect another volume
  /// \return the bounding volume of this object. Memory lifetime for this volume is
  ///         the responsibility of the primitive. Do not hold on to this pointer; create
  ///         a copy if you need to save this volume for some later use.
  ///         The function returns NULL if the primitive has no bounding volume
  virtual const UWB_BoundingVolume* GetBoundingVolume( eLevelOfDetail lod ) const;

  /// This function is called if it is determined that this primitive collides with another
  /// primitive. It is up to the primitive to perform some sort of response active when
  /// a collision occurs
  /// \param pOther primitive that this primitive collides with
  /// \param other_location location of other primitive
  virtual void CollisionResponse(const UWB_Primitive* pOther, vec3 other_location);

  /// Draw the bounding volume of this primitive as a wireframe
  /// \param lod the level of detail to use for drawing the bounding volume
  /// \param pRenderer the draw helper class to use for drawing functions
  /// \param color the color to draw the bounding volume
  void DrawBoundingVolume( eLevelOfDetail lod, UWB_DrawHelper* pRenderer, UWB_Color color ) const;


protected:
  //The DrawPrimitive and SetupDrawAttributes functions are used by the Draw function
  
  /// set the draw attributes on the passed in drawhelper class
  /// or API specific attributes that are not part of the drawhelper
  /// if this is a API specific primitive subclass
  virtual void SetupDrawAttributes(UWB_DrawHelper& draw_helper);

  /// Subclasses must implement a DrawPrimitive function
  /// \param draw_helper Drawing operations are called on the draw_helper in
  ///        order to abstract the graphics API from the primitive class
  virtual void DrawPrimitive( eLevelOfDetail lod, UWB_DrawHelper& draw_helper ) = 0;

public:
  // Attributes
  // basic object attributes - they may not make sense to all subclasses,
  // but they are commom and small enough that it is fine to provide them
  // for all objects

  ///
  /// Object visibility. If it is not visible, it will not get a call to DrawPrimitive
  void SetVisible(bool on);
  ///
  /// Object visibility. If it is not visible, it will not get a call to DrawPrimitive
  bool IsVisible() const;

  /// Velocity is the amount of distance traveled over an elapsed period of time. This
  /// by the Update function to adjust the position of a primitive
  void SetVelocity(vec3 velocity);
  /// Velocity is the amount of distance traveled over an elapsed period of time. This
  /// by the Update function to adjust the position of a primitive
  vec3 GetVelocity() const;
  ///
  /// Returns true if the primitive's velocity is {0,0,0}
  bool IsStationary() const;

  void SetFlatColor(UWB_Color color);
  UWB_Color GetFlatColor() const;
  void SetShadingColor(UWB_Color color);
  UWB_Color GetShadingColor() const;

  void SetFillMode( eFillMode mode );
  eFillMode GetFillMode() const;

  void SetShadeMode( eShadeMode mode );
  eShadeMode GetShadeMode() const;

  void SetPointSize(float size);
  float GetPointSize() const;

  void SetDrawBoundingVolume(bool on);
  bool IsDrawBoundingVolume() const;

  void EnableBlending( bool on );
  bool IsBlendingEnabled() const;

  void EnableLighting(bool on);
  bool IsLightingEnabled() const;

  UWB_Material GetMaterial() const;
  void SetMaterial(const UWB_Material&);

  void SetTextureFileName(const wchar_t* texFile);
  const wchar_t* GetTextureFileName();
  void EnableTexturing(bool on);
  bool IsTexturingEnabled() const;
protected:
  bool m_bVisible;            // default true
  vec3 m_velocity;            // default 0,0,0 (stopped)
  UWB_Color m_FlatColor;      // default black
  UWB_Color m_ShadingColor;   // default white
  eFillMode m_FillMode;       // default solid
  eShadeMode m_ShadeMode;     // default Gouraud
  float m_PointSize;          // default 2
  bool m_bDrawBoundingVolume; // default false
  bool m_bBlendingEnabled;
  bool m_bTexturingEnabled;
  bool m_bLightingEnabled;
  UWB_Material m_Material;
  std::wstring m_TexFileName;
};
